from typing import Dict
from flask import request
from functools import cached_property

from origin.api.context import Context


class FlaskContext(Context):
    """Flask-specific context."""

    @cached_property
    def headers(self) -> Dict[str, str]:
        """
        Handle the header.

        :returns: HTTP request headers
        """
        return dict(request.headers)

    @cached_property
    def cookies(self) -> Dict[str, str]:
        """
        Handle the cookie.

        :returns: HTTP request cookies
        """
        return request.cookies



from enum import Enum
from dataclasses import dataclass

from origin.serialize import Serializable


class TechnologyType(Enum):
    """System-wide labels of known technologies."""

    COAL = 'coal'
    NUCLEAR = 'nuclear'
    SOLAR = 'solar'
    WIND = 'wind'


@dataclass
class TechnologyCodes(Serializable):
    """
    Technology Code and Fuel Source Code for Electricity Production.

    Codes are described in the standard EECS Rules Fact Sheet 5: TYPES OF
    ENERGY INPUTS AND TECHNOLOGIES.
    """

    tech_code: str
    fuel_code: str


@dataclass
class Technology(Serializable):
    """
    Technology description from the standard EECS Rules Fact Sheet 5.

    A technology described by the standard described in the EECS Rules Fact
    Sheet 5: TYPES OF ENERGY INPUTS AND TECHNOLOGIES.
    """

    tech_code: str
    fuel_code: str
    type: TechnologyType



from dataclasses import dataclass

from origin.bus import Message
from origin.models.residual_mix import \
    ResidualMixEmissions, ResidualMixTechnologyDistribution


@dataclass
class ResidualMixEmissionsUpdate(Message):
    """
    Description of emissions.

    Describes emissions in the general mix within a single sector (pricing
    area), and in a specific time-frame.
    """

    emissions: ResidualMixEmissions


@dataclass
class ResidualMixTechnologyDistributionUpdate(Message):
    """
    Distribution of technologies.

    Describes the distribution of technologies used to produce energy
    in the residual mix within a single sector (pricing area),
    and in a specific time-frame.
    """

    distributions: ResidualMixTechnologyDistribution



from dataclasses import dataclass

from origin.bus import Message
from origin.models.tech import Technology, TechnologyCodes


@dataclass
class TechnologyUpdate(Message):
    """A Technology has been added or updated."""

    technology: Technology


@dataclass
class TechnologyRemoved(Message):
    """A Technology has been removed."""

    codes: TechnologyCodes



from dataclasses import dataclass

from origin.bus import Message
from origin.models.common import DateTimeRange
from origin.models.measurements import Measurement


@dataclass
class MeasurementUpdate(Message):
    """A new Measurement has been added to the system."""

    measurement: Measurement


@dataclass
class MeasurementRemoved(Message):
    """A Measurement has been removed from the system."""

    id: str


@dataclass
class ImportMeasurements(Message):
    """A Measurement has been imported to the system.

    begin.from_ is INCLUDED.
    begin.to_ is EXCLUDED.
    """

    gsrn: str
    begin: DateTimeRange



from dataclasses import dataclass

from origin.bus import Message
from origin.models.delegates import MeteringPointDelegate


@dataclass
class MeteringPointDelegateGranted(Message):
    """
    An actor has been delegated access to a MeteringPoint.

    An actor is identified by its subject
    """

    delegate: MeteringPointDelegate


@dataclass
class MeteringPointDelegateRevoked(Message):
    """
    An actor has had its delegated access to a MeteringPoint revoked.

    An actor is identified by its subject.
    """

    delegate: MeteringPointDelegate



from sqlalchemy import orm
from abc import abstractmethod


class SqlQuery(object):
    """ORM-level SQL construction class."""

    def __init__(self, session: orm.Session, query: orm.Query = None):
        self.session = session
        self.query = query or self._get_base_query()

    @abstractmethod
    def _get_base_query(self) -> orm.Query:
        """Handle the error if the Query is bad."""
        raise NotImplementedError

    def __iter__(self):
        """Iterate though the query."""

        return iter(self.query)

    def __getattr__(self, name):
        """Get the name attribute for the object."""

        return getattr(self.query, name)

    def filter(self, *filters):
        """
        Apply the given filtering criterion (keyword) to a copy of the Query.

        Example usage::
            filter(Model.age==35, Model.country=='Denmark')

        :param filters: filtering criterion
        :return: Result of the filtering criterion
        """
        return self.__class__(self.session, self.query.filter(*filters))

    def filter_by(self, **filters):
        """
        Apply the given filtering criteria (keywords) to a copy of the Query.

        Example usage::
            filter_by(age=35, country='Denmark')

        :param filters: filtering criteria
        :return: Result of the filtering criteria
        """
        return self.__class__(self.session, self.query.filter_by(**filters))

    def only(self, *fields):
        """
        Narrows down the columns to select.

        TODO Example usage

        :param fields:
        :return:
        """
        return self.__class__(self.session, self.query.options(
            orm.load_only(*fields)
        ))

    def get(self, field):
        """
        TODO.

        TODO Example usage

        :param field:
        :return: value for the field from the first result.
        """
        return self.only(field).scalar()

    def exists(self):
        """
        TODO.

        :return: True if result count is >= 1
        """
        return self.count() > 0



from typing import List
from abc import abstractmethod

from .context import Context
from .responses import Unauthorized


class EndpointGuard(object):
    """Only Allows requests with a valid endpoint provided."""

    @abstractmethod
    def validate(self, context: Context):
        """Validate the endpoint."""
        raise NotImplementedError


class TokenGuard(EndpointGuard):
    """Only Allows requests with a valid token provided."""

    def validate(self, context: Context):
        """Validate the token."""
        if context.token is None:
            raise Unauthorized('Unauthorized')


class ScopedGuard(EndpointGuard):
    """Only Allows requests with specific scopes granted."""

    def __init__(self, *scopes: str):
        self.scopes = scopes

    def validate(self, context: Context):
        """Validate the scope."""
        if context.token is None:
            raise Unauthorized('Unauthorized')
        for scope in self.scopes:
            if scope not in context.token.scope:
                # TODO Write proper message
                raise Unauthorized('Missing scope %s' % scope)


class Bouncer(object):
    """Class Bouncer to validate the context and the endpoints."""

    def validate(self, context: Context, guards: List[EndpointGuard]):
        """Validate the bouncer."""
        for guard in guards:
            guard.validate(context)


# -- Singletons --------------------------------------------------------------

bouncer = Bouncer()



from typing import Dict, Any, Optional
from urllib.parse import urlencode
import urllib.parse as urlparse


def url_append(
    url: str,
    query_extra: Optional[Dict[str, Any]] = None,
    path_extra: Optional[str] = None,
) -> str:
    """
    Change the provided URL, add or overwrite.

    Provided a URL, this function adds (or overwrites) one or more query
    parameters and/or path, and returns the updated URL.

    :param url: The URL to add query parameters to
    :param query_extra: The query parameters to add
    :param path_extra: The path to add
    :return: The URL with added query parameters
    """
    url_parts = list(urlparse.urlparse(url))

    if path_extra is not None:
        url_parts[2] += path_extra

    if query_extra is not None:
        query = dict(urlparse.parse_qsl(url_parts[4]))
        query.update(query_extra)
        url_parts[4] = urlencode(query)

    return urlparse.urlunparse(url_parts)



from ..endpoint import Endpoint


class OpenApiSchema(Endpoint):
    """OpenAPI Schema endpoint."""

    def __init__(self, app):
        self.app = app

    def handle_request(self):
        """Handle request."""

        pass



from dataclasses import dataclass

from origin.serialize import Serializable


@dataclass
class MeteringPointDelegate(Serializable):
    """
    Delegation of meteringPoints to actors.

    An actor (identified by its subject) who has been delegated
    access to a MeteringPoint (identified by its GSRN number).
    """

    subject: str
    gsrn: str

    # TODO Define time period (???)



from typing import Optional
from urllib.parse import urlparse, parse_qs


def assert_base_url(
        url: str,
        expected_base_url: str,
        check_path: bool = True,
):
    """
    Assert to test the base url.

    Asserts that a given [absolute] URL has the expected base URL,
    ie. the protocol and domain parts.

    Example of tests that will pass::

        assert_base_url('http://foo.com', 'http://foo.com')
        assert_base_url('https://foo.com?query=value', 'https://foo.com')

    Examples of tests that will fail::

        assert_has_base_url('http://bar.com', 'http://bar.com')
        assert_has_base_url('https://bar.com?query=value', 'https://bar.com')

    :param url: The URL to test
    :param expected_base_url: The base-URL to match
    :param check_path: Whether or not the assert on the path also
    """

    url_parsed = urlparse(url)
    expected_url_parsed = urlparse(expected_base_url)

    error_msg = f'URL mismatch: {url} != {expected_base_url}'

    assert url_parsed.scheme == expected_url_parsed.scheme, error_msg
    assert url_parsed.netloc == expected_url_parsed.netloc, error_msg

    if check_path:
        assert url_parsed.path == expected_url_parsed.path, error_msg
        assert url_parsed.params == expected_url_parsed.params, error_msg


def assert_query_parameter(
        url: str,
        name: str,
        value: Optional[str] = None,
):
    """
    Assert to test the query parameters.

    Asserts that a given [absolute] URL has the expected query parameter
    present. Optionally asserts the value of the parameter, otherwise only
    assert the existence of the parameter.

    Example of tests that will pass::

        assert_query_parameter('http://foo.com?query=value', 'query')
        assert_query_parameter('http://foo.com?query=value', 'query', 'value')

    Examples of tests that will fail::

        assert_query_parameter('http://bar.com?query=value', 'foo')
        assert_query_parameter('http://bar.com?query=value', 'query', 'bar')

    :param url: The URL to test
    :param name: Name of the query parameter
    :param value: Value of the query parameter to match, or None
    """

    url_parsed = urlparse(url)
    query_parsed = parse_qs(url_parsed.query)

    assert name in query_parsed, \
        f'Query parameter "{name}" not found in: {query_parsed}'

    if value is not None:
        assert query_parsed[name] == [value], (
            f'Value for query parameter {name} mismatch: '
            f'{query_parsed[name]} != {value}'
        )
